use std;
use std::collections::HashMap;
use std::slice::Iter;
use std::vec::IntoIter;
use std::iter::Zip;
use clusterable::Clusterable;

pub struct Kmeans<T: Clusterable> {
    centroids: Vec<T>,
    elements: Vec<T>,
    labels: Vec<usize>,
    cluster_number: usize,
}

impl<T:Clusterable> Kmeans<T> {
    pub fn new(centroids: Vec<T>, data: Vec<T>) -> Kmeans<T> {
        let labels = Kmeans::build_labels(&centroids, &data);
        let cluster_number = centroids.len();
        Kmeans {
            centroids: centroids,
            elements: data,
            labels: labels,
            cluster_number: cluster_number
        }
    }

    /// \returns True if converged
    pub fn iterate(&mut self) -> bool {
        // Update the centroids
        let centroids = Kmeans::build_centroids(&self.elements, &self.labels, self.cluster_number);

        if self.centroids == centroids {
            true
        } else {
            self.centroids = centroids;
            Kmeans::update_labels(&self.centroids, &self.elements, &mut self.labels);
            false
        }
    }

    pub fn build_labels(centroids: &Vec<T>, data: &Vec<T>) -> Vec<usize> {
        debug_assert_ne!(0, centroids.len());

        let mut output = vec![0; data.len()];
        Kmeans::update_labels(centroids, data, &mut output);
        output
    }

    pub fn update_labels(centroids: &Vec<T>, data: &Vec<T>, labels: &mut Vec<usize>) {
        for (element, new_label) in data.iter().zip(labels.iter_mut()) {
            *new_label = centroids
                .iter()
                .enumerate()
                .min_by(|&(_, c1), &(_, c2)| {
                    c1.distance(element).partial_cmp(&c2.distance(element)).unwrap()
                }).unwrap().0;
        }
    }

    pub fn build_centroids(data: &Vec<T>, labels: &Vec<usize>, cluster_number: usize) -> Vec<T> {

        let mut centroids = vec![];
        for label in 0..cluster_number {

            let to_consider = data
                .iter()
                .enumerate()
                .filter(|&(index, _)| labels[index] == label)
                .map(|(_, element)| element);

            if let Some(centroid) = T::get_centroid(to_consider) {
                centroids.push(centroid);
            }
        }

        centroids
    }

    pub fn iter(&self) -> Zip<Iter<T>, Iter<usize>> {
        debug_assert_eq!(self.elements.len(), self.labels.len());
        return self.elements.iter().zip(self.labels.iter());
    }

    pub fn into_iter(self) -> Zip<IntoIter<T>, IntoIter<usize>> {
        debug_assert_eq!(self.elements.len(), self.labels.len());
        return self.elements.into_iter().zip(self.labels.into_iter());
    }

    pub fn into_vec_vec(self) -> Vec<Vec<T>> {
        let mut map = HashMap::new();

        for (element, label) in self.into_iter() {
            let mut entry = map.entry(label).or_insert(vec![]);
            entry.push(element);
        }

        let mut output = vec![];

        for (_, cluster) in map {
            let mut vec = vec![];
            for element in cluster {
                vec.push(element);
            }
            output.push(vec);
        }

        output
    }

    pub fn to_vec_vec(&self) -> Vec<(Vec<T>, usize)> {
        let mut map = HashMap::new();

        for (element, label) in self.iter() {
            let mut entry = map.entry(*label).or_insert(vec![]);
            entry.push(element.clone());
        }

        let mut output = vec![];

        for (label, cluster) in map {
            let mut vec = vec![];
            for element in cluster {
                vec.push(element);
            }
            output.push((vec, label));
        }

        output
    }

}

pub enum Error {
    IterationsLimitExceeded,
}

pub fn kmeans<T: Clusterable>(centroids: Vec<T>, data: Vec<T>, max_iterations: usize)
    -> Result<(Kmeans<T>, usize), Error> {

    let mut kmeans = Kmeans::new(centroids, data);

    for nb_iterations in 0..max_iterations {

        let stable = kmeans.iterate();

        if stable {
            return Ok((kmeans, nb_iterations));
        }

    }

    Err(Error::IterationsLimitExceeded)
}

pub fn equal_kmeans<T: Clusterable>(centroids: Vec<T>, data: Vec<T>, max_iterations: usize)
    -> Result<(Kmeans<T>, usize), Error> {

    let number_of_elements = data.len();
    let (kmeans, nb_iterations) = kmeans(centroids, data, max_iterations)?;
    let mut clusters = kmeans.to_vec_vec();

    for &mut (ref mut cluster, ref mut label) in &mut clusters {
        cluster.sort_by(|ref e1, ref e2| {
            *&e1.distance(&kmeans.centroids[*label])
                .partial_cmp(&e2.distance(&kmeans.centroids[*label])).unwrap()
        });
    }

    let max = (number_of_elements as f64 / clusters.len() as f64).ceil() as usize;

    let mut to_replace = vec![];
    for &mut (ref mut cluster, _) in &mut clusters {
        while cluster.len() > max {
            to_replace.push(cluster.pop().unwrap());
        }
    }

    for element in to_replace {
        // Find the best non full cluster to place it in
        let mut best_distance = std::f64::MAX;
        let mut best_index = 0;

        for (index, centroid) in kmeans.centroids.iter().enumerate() {

            let new_distance = element.distance(centroid);
            if new_distance < best_distance && clusters[index].0.len() < max {
                best_distance = new_distance;
                best_index = index;
            }

        }

        clusters[best_index].0.push(element);
    }

    let mut elements = vec![];
    let mut labels = vec![];
    let mut centroids = vec![];

    for (index, &(ref cluster, _)) in clusters.iter().enumerate() {
        for element in cluster {
            elements.push(element.clone());
            labels.push(index);
        }
        centroids.push(Clusterable::get_centroid(cluster.iter()).unwrap());
    }

    let len = centroids.len();

    // Build a new k-means from clusters
    let new_kmeans = Kmeans {
        centroids: centroids,
        elements: elements,
        labels: labels,
        cluster_number: len,
    };

    Ok((new_kmeans, nb_iterations))

}
